@@ -1,5 +1,5 @@
 from collections import defaultdict, ChainMap
-from datetime import date
+from datetime import date, datetime, timezone
 from contextlib import contextmanager
 
 from typing import Optional, Dict, Sequence, Mapping, NamedTuple, Set
@@ -7,6 +7,7 @@ import typing as T
 from warnings import warn
 
 import jinja2
+from jinja2 import nativetypes
 import yaml
 
 from .utils.template_utils import FakerTemplateLibrary
@@ -16,7 +17,8 @@ from .template_funcs import StandardFuncs
 from .data_gen_exceptions import DataGenSyntaxError, DataGenNameError
 import snowfakery  # noQA
 from snowfakery.object_rows import NicknameSlot, SlotState, ObjectRow
-from snowfakery.plugins import PluginContext, SnowfakeryPlugin
+from snowfakery.plugins import PluginContext, SnowfakeryPlugin, ScalarTypes
+from snowfakery.utils.collections import OrderedSet
 
 OutputStream = "snowfakery.output_streams.OutputStream"
 VariableDefinition = "snowfakery.data_generator_runtime_object_model.VariableDefinition"
@@ -120,7 +122,7 @@ class Globals:
         self.persistent_objects_by_table = {}
 
         self.id_manager = IdManager()
-        self.intertable_dependencies = set()
+        self.intertable_dependencies = OrderedSet()
         self.today = today or date.today()
         self.nicknames_and_tables = name_slots or {}
 
@@ -218,7 +220,7 @@ class Globals:
         self.nicknames_and_tables = state["nicknames_and_tables"]
         self.id_manager = hydrate(IdManager, state["id_manager"])
 
-        self.intertable_dependencies = set(
+        self.intertable_dependencies = OrderedSet(
             Dependency(*dep) for dep in getattr(state, "intertable_dependencies", [])
         )
 
@@ -233,7 +235,21 @@ class Globals:
 
 
 class JinjaTemplateEvaluatorFactory:
-    def __init__(self):
+    def __init__(self, native_types: bool):
+        if native_types:
+            self.compilers = [
+                nativetypes.NativeEnvironment(
+                    block_start_string="${%",
+                    block_end_string="%}",
+                    variable_start_string="${{",
+                    variable_end_string="}}",
+                )
+            ]
+
+            return
+
+        # TODO: Delete this old code_path when the
+        #       transition to native_types is complete.
         self.compilers = [
             jinja2.Environment(
                 block_start_string="${%",
@@ -323,6 +339,15 @@ class Interpreter:
         self.statements = parse_result.statements
         self.parent_application = parent_application
         self.instance_states = {}
+        self.filter_row_values = self.filter_row_values_normal
+        snowfakery_version = self.options.get(
+            "snowfakery.standard_plugins.SnowfakeryVersion.snowfakery_version", 2
+        )
+        assert snowfakery_version in (2, 3)
+        self.native_types = snowfakery_version == 3
+        self.template_evaluator_factory = JinjaTemplateEvaluatorFactory(
+            self.native_types
+        )
 
     def execute(self):
         self.current_context = RuntimeContext(interpreter=self)
@@ -412,6 +437,14 @@ class Interpreter:
             self.instance_states[uniq_name] = [parent_obj, value]
         return value
 
+    def filter_row_values_normal(self, row: dict):
+        return {k: v for k, v in row.items() if not k.startswith("__")}
+
+    # for future use:
+
+    # def filter_row_values_and_remove_ids(self, row: dict):
+    #     return {k: v for k, v in row.items() if not k.startswith("__") or k == "id"}
+
 
 class RuntimeContext:
     """Local "stack frame" type object. RuntimeContexts live on the Python stack.
@@ -422,9 +455,10 @@ class RuntimeContext:
     but internally its mostly just proxying to other classes."""
 
     obj: Optional[ObjectRow] = None
-    template_evaluator_recipe = JinjaTemplateEvaluatorFactory()
     current_template = None
     local_vars = None
+    unique_context_identifier = None
+    recalculate_every_time = False
 
     def __init__(
         self,
@@ -440,12 +474,17 @@ class RuntimeContext:
         self.parent = parent_context
         if self.parent:
             self._plugin_context_vars = self.parent._plugin_context_vars.new_child()
+            self.recalculate_every_time = parent_context.recalculate_every_time
         else:
             self._plugin_context_vars = ChainMap()
         locale = self.variable_definitions().get("snowfakery_locale")
         self.faker_template_library = self.interpreter.faker_template_library(locale)
         self.local_vars = {}
 
+    @property
+    def filter_row_values(self):
+        return self.interpreter.filter_row_values
+
     # TODO: move this into the interpreter object
     def check_if_finished(self):
         "Have we iterated over the script enough times?"
@@ -508,7 +547,7 @@ class RuntimeContext:
         return self.interpreter.output_stream
 
     def get_evaluator(self, definition: str):
-        return self.template_evaluator_recipe.get_evaluator(definition)
+        return self.interpreter.template_evaluator_factory.get_evaluator(definition)
 
     @property
     def evaluation_namespace(self):
@@ -549,6 +588,7 @@ class EvaluationNamespace(NamedTuple):
             "child_index": obj._child_index if obj else None,
             "this": obj,
             "today": interpreter.globals.today,
+            "now": datetime.now(timezone.utc),
             "fake": self.runtime_context.faker_template_library,
             "template": self.runtime_context.current_template,
             **interpreter.options,
@@ -568,7 +608,11 @@ class EvaluationNamespace(NamedTuple):
         return {**self.field_funcs(), "fake": self.fake}
 
     def fake(self, name):
-        return str(self.runtime_context.faker_template_library._get_fake_data(name))
+        val = self.runtime_context.faker_template_library._get_fake_data(name)
+        if isinstance(val, ScalarTypes):
+            return val
+        else:
+            return str(val)
 
     def field_vars(self):
         return {**self.simple_field_vars(), **self.field_funcs()}
