@@ -5,9 +5,11 @@ import csv
 import subprocess
 import datetime
 import sys
+from decimal import Decimal
 from pathlib import Path
 from collections import namedtuple, defaultdict
 from typing import Dict, Union, Optional, Mapping, Callable, Sequence
+import typing as T
 from warnings import warn
 
 from sqlalchemy import (
@@ -35,6 +37,11 @@ def noop(x):
     return x
 
 
+def format_datetime(dt: datetime.datetime):
+    """Format into the Salesforce-preferred syntax."""
+    return dt.isoformat(timespec="seconds")
+
+
 class OutputStream(ABC):
     """Common base class for all output streams"""
 
@@ -46,9 +53,10 @@ class OutputStream(ABC):
         int: int,
         float: float,
         datetime.date: noop,
-        datetime.datetime: noop,
+        datetime.datetime: format_datetime,
         type(None): noop,
         bool: int,
+        Decimal: str,
     }
     uses_folder = False
     uses_path = False
@@ -92,17 +100,12 @@ class OutputStream(ABC):
                 )
             return encoder(field_value)
 
-    def should_output(self, value):
-        return not value.startswith("__")
-
     def write_row(self, tablename: str, row_with_references: Dict) -> None:
-        should_output = self.should_output
         row_cleaned_up_and_flattened = {
             field_name: self.cleanup(
                 field_name, field_value, tablename, row_with_references
             )
             for field_name, field_value in row_with_references.items()
-            if should_output(field_name)
         }
         self.write_single_row(tablename, row_cleaned_up_and_flattened)
         if self.count % self.flush_limit == 0:
@@ -118,7 +121,7 @@ class OutputStream(ABC):
         """Write a single row to the stream"""
         pass
 
-    def close(self) -> Optional[Sequence[str]]:
+    def close(self, **kwargs) -> Optional[Sequence[str]]:
         """Close any resources the stream opened.
 
         Do not close file handles which were passed in!
@@ -143,7 +146,7 @@ class SmartStream:
 
     mode = "wt"
 
-    def __init__(self, stream_or_path=None):
+    def __init__(self, stream_or_path=None, **kwargs):
         if hasattr(stream_or_path, "write"):
             self.owns_stream = False
             self.stream = stream_or_path
@@ -159,7 +162,7 @@ class SmartStream:
     def write(self, data):
         self.stream.write(data)
 
-    def close(self):
+    def close(self, **kwargs):
         if self.owns_stream:
             self.stream.close()
             return [f"Generated {self.stream.name}"]
@@ -173,7 +176,7 @@ class FileOutputStream(OutputStream):
         self.write = self.smart_stream.write
         self.stream = self.smart_stream.stream
 
-    def close(self):
+    def close(self, **kwargs):
         return self.smart_stream.close()
 
 
@@ -211,7 +214,7 @@ class CSVOutputStream(OutputStream):
             Path.mkdir(self.target_path, exist_ok=True)
 
     def open_writer(self, table_name, table):
-        file = open(self.target_path / f"{table_name}.csv", "w")
+        file = open(self.target_path / f"{table_name}.csv", "w", newline="")
         writer = csv.DictWriter(file, list(table.fields.keys()) + ["id"])
         writer.writeheader()
         return CSVContext(dictwriter=writer, file=file)
@@ -225,7 +228,7 @@ class CSVOutputStream(OutputStream):
     def write_single_row(self, tablename: str, row: Dict) -> None:
         self.writers[tablename].dictwriter.writerow(row)
 
-    def close(self) -> Optional[Sequence[str]]:
+    def close(self, **kwargs) -> Optional[Sequence[str]]:
         messages = []
         for context in self.writers.values():
             context.file.close()
@@ -250,6 +253,7 @@ class JSONOutputStream(FileOutputStream):
         **OutputStream.encoders,
         datetime.date: str,
         datetime.datetime: str,
+        bool: bool,
     }
     is_text = True
 
@@ -267,7 +271,7 @@ class JSONOutputStream(FileOutputStream):
         values = {"_table": tablename, **row}
         self.write(json.dumps(values))
 
-    def close(self) -> Optional[Sequence[str]]:
+    def close(self, **kwargs) -> Optional[Sequence[str]]:
         if not self.first_row:
             self.write("]\n")
         return super().close()
@@ -326,7 +330,7 @@ class SqlDbOutputStream(OutputStream):
         self.flush()
         self.session.commit()
 
-    def close(self) -> Optional[Sequence[str]]:
+    def close(self, **kwargs) -> Optional[Sequence[str]]:
         self.commit()
         self.session.close()
 
@@ -340,14 +344,19 @@ class SqlDbOutputStream(OutputStream):
 
         for tablename, model in self.metadata.tables.items():
             if tablename in inferred_tables:
-                self.table_info[tablename] = TableTuple(
+                table_info = TableTuple(
                     insert_statement=model.insert(bind=self.engine, inline=True),
                     fallback_dict={
                         key: None for key in inferred_tables[tablename].fields.keys()
                     },
                 )
                 # id is special
-                self.table_info[tablename].fallback_dict.setdefault("id", None)
+                table_info.fallback_dict.setdefault("id", None)
+
+                # See create_tables_from_inferred_fields to see what _sf_update_key are for
+                if inferred_tables[tablename].has_update_keys:
+                    table_info.fallback_dict.setdefault("_sf_update_key", None)
+                self.table_info[tablename] = table_info
 
 
 # backwards-compatible name for CCI
@@ -402,7 +411,9 @@ class SqlTextOutputStream(FileOutputStream):
         self.tempdir.cleanup()
 
 
-def create_tables_from_inferred_fields(tables, engine, metadata):
+def create_tables_from_inferred_fields(
+    tables: T.Dict[str, TableInfo], engine, metadata
+):
     """Create tables based on dictionary of tables->field-list."""
     with engine.connect() as conn:
         inspector = inspect(engine)
@@ -421,6 +432,12 @@ def create_tables_from_inferred_fields(tables, engine, metadata):
                     "id", Integer(), primary_key=True, autoincrement=True
                 )
 
+            # add a column keeping track of what update_key was specified by
+            # the template. This allows multiple templates to have different
+            # update_keys.
+            if table.has_update_keys:
+                columns.append(Column("_sf_update_key", Unicode(255)))
+
             t = Table(table_name, metadata, id_column, *columns)
 
             if inspector.has_table(table_name):
@@ -493,7 +510,7 @@ class GraphvizOutputStream(FileOutputStream):
         )
         self.nodes[tablename, row["id"]] = self.G.newItem(node_name)
 
-    def close(self) -> Optional[Sequence[str]]:
+    def close(self, **kwargs) -> Optional[Sequence[str]]:
         for fieldname, source, target in self.links:
             mylink = self.G.newLink(self.nodes[source], self.nodes[target])
             self.G.propertyAppend(mylink, "label", fieldname)
@@ -539,7 +556,7 @@ class ImageOutputStream(OutputStream):
     def flatten(self, *args, **kwargs):
         return self.gv_os.flatten(*args, **kwargs)
 
-    def close(self) -> Optional[Sequence[str]]:
+    def close(self, **kwargs) -> Optional[Sequence[str]]:
         self.gv_os.close()
         assert self.dotfile.exists()
         rc = self._render(self.dotfile, self.path)
@@ -577,7 +594,7 @@ class MultiplexOutputStream(OutputStream):
         for stream in self.outputstreams:
             stream.write_row(tablename, row_with_references)
 
-    def close(self) -> Optional[Sequence[str]]:
+    def close(self, **kwargs) -> Optional[Sequence[str]]:
         for stream in self.outputstreams:
             stream.close()
 
